#include "cpp_config.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <xmmintrin.h>
#include <emmintrin.h>
#ifdef HAS_SSE41 
#include <smmintrin.h>
#endif
#ifdef HAS_SSE4A 
#include <smmintrin.h>
#endif

#include <immintrin.h>


#if defined(HAS_AVX) || defined(HAS_AVX2)
#define ALIGN_BYTES 32
#else
#define ALIGN_BYTES 16
#endif

#ifdef HAS_POPCNT
#define POPCNT(x) _mm_popcnt_u32(x)
#else
#define POPCNT(x) __builtin_popcount(x)
#endif


/**
 * definition of prototype of mesure function
 */
typedef float (*mesureFunction)(char *v, char *w, uint size);

/**
 * define here methods written in assembly code
 */

#ifdef HAS_SSE2
extern "C" float mesure_sse2(char *v, char *w, uint size); 
#endif

/**
 * reference function that needs to be improved
 * @param v array of size characters
 * @param w array of size characters
 * @param size size of arrays v and w
 * @return ratio identical / (identical + different)
 * !!!! DONT'T MODIFY THIS FUNCTION !!!!
 */
float mesure_reference(char *v, char *w, uint size) {
	uint i, s, d;
	
	s = d = 0;
	for (i = 0; i <size; ++i) {
		if (v[i] == w[i]) ++s; else ++d;
	}
	return ((float) s / (float)(s + d));
}


typedef struct FnDescriptor {
	mesureFunction func;
	const char *name;
} FnDescriptor;


/**
 * reference your functions here with there names
 *
 */
FnDescriptor tab_methods[] = {
	{ mesure_reference, "reference" },
#ifdef HAS_SSE2
	{ mesure_sse2,		"sse2 asm"},
#endif	
	{ NULL, "" }
};



// ==================================================================
// ==================================================================
// DONT MODIFY THE FOLLOWING !!!!!!!!!!!!!!!!
// ==================================================================
// ==================================================================

/**
 * return number of methods/functions
 */
int nbr_methods() {
	int i=0;
	while (tab_methods[i].func != NULL) ++i;
	return i;
}

void report_methods_names() {
	int i=0;
	while (tab_methods[i].func != NULL) {
		printf("method %d=%s\n", i, tab_methods[i].name);
		++i;
	}
}

int M = 0;		// method used 0 = default
int K = 100;	// number of sequences or strings
int L = 100;	// length of sequences >= 16
int R = 10000;	// number of repeat

// pointer to function
mesureFunction mesure = &mesure_reference;

// data
char **sequences;

/**
 * allocate K sequences of length L and setup
 *
 */
void allocate_and_set_memory() {
	int k, l;

	int alignment = ALIGN_BYTES;

	sequences = (char **) _mm_malloc(K * sizeof(char *), alignment); 
	for (k=0; k<K; ++k) {
		char *s = (char *) _mm_malloc(L, alignment);
		sequences[k] = s;
		for (l=0; l<L; ++l) s[l] = (rand() % 5) + 65;
	}
}

/**
 * test all methods available, we consider that method 0 (default)
 * gives the exact result
 * report error if method doesn't report the exact result
 */
void test() {
	int i = 0;
	const float error = 1e-6;
	
	float expected_result = mesure_reference(sequences[0], sequences[1], L);
	
	++i;
	while (tab_methods[i].func != NULL) {
		float result = (*tab_methods[i].func)(sequences[0], sequences[1], L);
		/* printf("> %f\n", result); */
		if (fabs(result - expected_result) > error) {
			printf("error: bad value for method %d, found %f instead of %f\n",
				i, result, expected_result);
		}
		++i;
	}
}

/**
 * perform main computation that will be evaluated
 * we call the mesure function R * K * (K+1) / 2 times
 */
void compute() {
	int i, j, r;
	
	for (r = 0; r < R; ++r) {
		float sum = 0.0;
		for (i=0; i<K-1; ++i) {
			for (j=i+1; j<K; ++j) {
				sum += (*mesure)(sequences[i], sequences[j], L);
			}
		}
		if (r==0) printf("sum = %f\n", sum);
	}
}

/**
 * main method
 * accepts parameters :
 * first (if present) : method:
 *  -- if -1 perform test of each method
 *  -- if -2 report number of methods and method names
 *  -- if >= 0 use method
 * second (if present) : length of sequences (default value 100)
 * third (if present) : number of sequences K (default value 100)
 */
int main(int argc, char *argv[]) {
	srand(33);
	
	if (argc > 1) {
		M = atoi(argv[1]);
	}
	if (argc > 2) {
		L = atoi(argv[2]);
	}	
	if (argc > 3) {
		K = atoi(argv[3]);
	}
	
	allocate_and_set_memory();

	if (M == -1) {
		test();
	} else if (M == -2) {
		printf("number of methods=%d\n", nbr_methods());
		printf("=====================\n");
		report_methods_names();	
	} else {	
		mesure = tab_methods[M].func;
		compute();
	}
	return 0;
}
